""" Compute Truck Factor

The truck factor assignment includes several factors. We're just
going to focus on the actual truck factor computation from a database
in this file.

References:
  Assignment: http://janvitek.org/events/NEU/6050/a4.html
  Paper: http://janvitek.org/events/NEU/6050/Ps/truck.pdf
"""
from collections import defaultdict

from sqlalchemy import func
from okra.models import (Meta, Author, Contrib, CommitFile, Info)

def total_number_of_files_by_project(owner, project, dal):
    """ Compute the total number of files by project. 
    
    :param owner: name of owner
    :param project: name of project
    :param dal: okra.models.DataAccessLayer
    :return: the number of files in a project
    :rtype: int
    """
    q = dal.session.query(
        Meta.owner_name, Meta.project_name, CommitFile.modified_file
    ).\
    join(CommitFile).\
    filter(
        Meta.owner_name == owner, Meta.project_name == project
    ).\
    group_by(CommitFile.modified_file).all()
    return len(q)

def total_number_of_contributors_by_project(owner, project, dal):
    """ Compute the total number of contributors by project.

    :param owner: name of owner
    :param project: name of project
    :param dal: okra.models.DataAccessLayer
    :return: the total number of contributors in a project
    :rtype: int
    """
    q = dal.session.query(
        Meta.owner_name, Meta.project_name, Author.email
    ).\
    filter(
        Meta.owner_name == owner, Meta.project_name == project
    ).\
    join(Author).\
    group_by(Author.name)
    results = q.all()
    return len(results)

def author_file_owned(owner, project, dal):
    """ Compute file ownership by each author. 

    :param owner: name of owner
    :param project: name of project    
    :param dal: okra.models.DataAccessLayer
    :return: list of file owners based on max number of lines written
    :rtype: list of sqlalchemy objects
    """

    # Number of lines added by each author per file
    
    q = dal.session.\
        query(
            Meta.owner_name, Meta.project_name, Meta.commit_hash,
            Author.name, Author.email, CommitFile.modified_file,
            func.sum(CommitFile.lines_added).label("total_lines_added")
        ).\
        filter(Meta.owner_name == owner, Meta.project_name == project).\
        join(Author).join(CommitFile).\
        group_by(Author.name, CommitFile.modified_file)

    # Author with max number of lines added per file (owner)

    res = q.all()

    owner = {}
    for item in res:

        fname = item.modified_file
        if fname in owner:

            if owner[fname].total_lines_added > item.total_lines_added:
                owner[fname] = item

        else:
            owner[fname] =  item

    results = [i for i in owner.values()]
    return results

def author_number_of_files_owned(results):
    """ Number of files owned by author. 

    :param results: results from author_file_owned() 
    :return: {author:  number of files owned}
    :rtype: dict
    """
    authors = defaultdict(int)
    for item in results:

        authors[item.name] += 1

    return authors

def smallest_owner_set(authors, total, size=0.5):
    """ Smallest set of authors owning more than half of project files. 

    :param authors: author_number_of_files_owned() output
    :param total: total_number_of_files_by_project() output
    :return: (number of members in smallest set, smallest set)
    :rtype: tuple
    """
    items = sorted(authors.items(), key=lambda k_v: k_v[1], reverse=True)
    threshold = round(total * size)
    agg = 0
    members = []
    member_count = 0
    for k,v in items:

        if agg <= threshold:

            members.append((k,v))
            member_count += 1
            agg += v

        else:
            break

    return (member_count, members)

def get_truck_factor_by_project(owner, project, dal):
    """ Get the 'truck factor' by project.

    1. For each project, and each file, compute how many lines were 
       added by each unique user.
    2. For each project, and each file, find which user created the file.
    3. Given the above two results compute the ownership of each file.
    4. For each project, and each file pick an owner.
    5. For each project, rank the users by the number of files they own.
    6. Given all of the above compute the Truck Factor as the smallest set 
       of users such that they own more than half of the files in the project

    :param owner: name of owner
    :param project: name of project
    :param dal: okra.models.DataAccessLayer
    :return: Truck factor score for a GitHub project, Truck set members
    :rtype: tuple (int, list)
    """
    total_files = total_number_of_files_by_project(owner, project, dal)
    owners = author_file_owned(owner, project, dal)
    authors = author_number_of_files_owned(owners)
    member_count, members = smallest_owner_set(authors, total_files)

    return (member_count, members)



""" SQL Database Models 

This is the database schema used for accessing the SQL database.
"""
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy import (Table, Column, Integer, String, MetaData, Numeric,
                        ForeignKey, DateTime, create_engine)

Base = declarative_base()


class Inventory(Base):

    __tablename__ = 'inventory'

    owner_name = Column('owner_name', String(100), nullable=False,
                        primary_key=True, index=True)
    project_name = Column('project_name', String(150), nullable=False,
                        primary_key=True, index=True)
    last_commit = Column('last_commit', String(40), nullable=True)

class Meta(Base):

    __tablename__ = 'meta'

    commit_hash = Column('commit_hash', String(40), 
                         ForeignKey('author.commit_hash'),
                         ForeignKey('contrib.commit_hash'),
                         ForeignKey('commit_file.commit_hash'),
                         ForeignKey('info.commit_hash'),
                         primary_key=True,
                         index=True)
    owner_name = Column('owner_name', String(100), nullable=False)
    project_name = Column('project_name', String(150), nullable=False)
    yearmo = Column('year_mo', String(7), nullable=False)

class Author(Base):
    """ 
    Author email is false, not all authors require a github account,
    so an email is not going to be required. 
    """
    __tablename__ = 'author'

    commit_hash = Column('commit_hash', String(40), primary_key=True,
                         index=True)
    name = Column('name', String(150), nullable=False,
                         index=True)
    email = Column('email', String(200), nullable=True)
    authored = Column('authored', DateTime, nullable=False)

class Contrib(Base):

    __tablename__ = 'contrib'
    contrib_id = Column('contrib_id', Integer, primary_key=True)
    commit_hash = Column('commit_hash', String(40), index=True,
                         nullable=False, primary_key=True)
    name = Column('name', String(150), index=True,
                          nullable=False)
    email = Column('email', String(200), nullable=True)
    contributed = Column('contributed', DateTime, nullable=False)

class CommitFile(Base):
    __tablename__ = 'commit_file'

    file_id = Column('file_id', Integer, primary_key=True)
    commit_hash = Column('commit_hash', String(40), index=True,
                         nullable=False,primary_key=True)
    modified_file = Column('modified_file', String(500), nullable=False)
    lines_added = Column('lines_added', Integer, nullable=False)
    lines_subtracted = Column('lines_subtracted', Integer, nullable=False)

class Info(Base):
    __tablename__ = 'info'

    commit_hash = Column('commit_hash', String(40), primary_key=True,
                         index=True)
    subject = Column('subject', String(250), nullable=True)
    message = Column('message', String(1500), nullable=True)
    created = Column('created', DateTime, nullable=False)

    
class DataAccessLayer(object):

    def __init__(self, conn_string):
        self.engine = None
        self.session = None
        self.conn_string = conn_string

    def connect(self):
        self.engine = create_engine(self.conn_string)
        Base.metadata.create_all(self.engine)
        self.Session = sessionmaker(bind=self.engine)




""" Populate SQL database. 

References:
  https://docs.sqlalchemy.org/en/latest/orm/tutorial.html
"""
import logging
from urllib.parse import urljoin

from okra.models import DataAccessLayer, Inventory, Meta
from okra.github import repo_to_objects
from okra.gitlogs import parse_inventory

logger = logging.getLogger(__name__)


def insert_buffer(items, dal, buffer_size=1024):
    """ Insert items using a buffer. 

    :param items: sqlalchemy orm database objects
    :param dal: okra.models.DataAccessLayer
    :param invobj: okra.models.Inventory object, to be updated
    :buffer_size: number of items to add before committing a session
    """
    logger.info("STARTED insert buffer")
    count = 0
    for item in items:
        
        dal.session.add(item)

        if count % buffer_size == 0:
            try:
                dal.session.commit()
                logger.info("Committed db objects: {}".format(count))
                
            except Exception as exc:
                dal.session.rollback()
                logger.error("Rolled back session")
                logger.exception(exc)
                raise exc

        count += 1

    try:
        dal.session.commit()
        logger.info("Committed db objects: {}".format(count))

    except Exception as exc:
        dal.session.rollback()
        logger.error("Rolled back session")
        logger.exception(exc)
        raise exc

    logger.info("COMPLETED insert buffer")

def persist_repo(owner:str, project:str, dburl:str, repopath:str, buffer_size:int):

    dal = DataAccessLayer(dburl)
    dal.connect()
    dal.session = dal.Session()

    last_commit = ""

    invobj = dal.session.query(
        Inventory.project_name, Inventory.owner_name,
        Inventory.last_commit).\
        filter(Inventory.project_name == project,
               Inventory.owner_name == owner).first() # unique value

    if invobj is not None:
        last_commit = invobj.last_commit

    # insert relevant commits into database

    objs = repo_to_objects(owner, project, repopath, last_commit)
    insert_buffer(objs, dal, buffer_size)

    # update inventory

    invmsg = parse_inventory(repopath, owner, project)

    if len(last_commit) > 0:

        # Retrieve new object b/c old object is no longer in session
            
        inv = dal.session.query(
            Inventory.project_name, Inventory.owner_name,
            Inventory.last_commit).\
            filter(Inventory.project_name == project,
                   Inventory.owner_name == owner).first() # unique value

        # Do not try updating database if object hasn't changed
            
        if inv.last_commit != invmsg.last_hash:
            inv.last_commit = invmsg.last_hash
            dal.session.add(inv)
            dal.session.commit()
    
    else:

        # Add new inventory object
            
        inv = Inventory(
            owner_name = invmsg.owner,
            project_name = invmsg.project,
            last_commit = invmsg.last_hash)

        dal.session.add(inv)
        dal.session.commit()

    dal.session.close()



""" Repository metrics for okra_api.proto """
import logging

from sqlalchemy import func

from okra.models import DataAccessLayer, Meta, Author, Contrib, CommitFile

logger = logging.getLogger(__name__)

def repo_info(dal: DataAccessLayer, yearmo=''):
    if len(yearmo) > 0:
        q = dal.session.query(
            Meta.yearmo, Contrib.commit_hash, Contrib.contributed
        ).join(Contrib).filter(Meta.yearmo == yearmo)
        return q.all()

    q = dal.session.query(Contrib.commit_hash, Contrib.contributed).\
        join(Contrib)
    return q.all()

def owner_info(dal: DataAccessLayer):
    q = dal.session.query(Meta.owner_name, Meta.project_name, Meta.commit_hash)
    return q.order_by(Meta.yearmo.desc()).first()

def iso_date_aggregation(dal: DataAccessLayer, yearmo=''):
    if len(yearmo) > 0:
        q = dal.session.query(
            Meta.commit_hash, Meta.yearmo, Contrib.contributed
        ).join(Contrib).filter(Meta.yearmo == yearmo)
        return q.all()

    q = dal.session.query(
        Meta.commit_hash, Meta.yearmo, Contrib.contributed
    ).join(Contrib)
    return q.all()

def find_author_count(dal: DataAccessLayer, yearmo=''):
    if len(yearmo) > 0:
        q = dal.session.query(
            Meta.commit_hash, Meta.yearmo, Contrib.name, Contrib.email
        ).join(Contrib).filter(Meta.yearmo == yearmo)
        return q.count()

    q = dal.session.query(
        Meta.commit_hash, Contrib.name, Contrib.email
    ).join(Contrib)
    return q.count()

def find_contributor_count(dal: DataAccessLayer, yearmo=''):
    if len(yearmo) > 0:
        q = dal.session.query(
            Meta.commit_hash, Meta.yearmo, Contrib.name, Contrib.email
        ).join(Contrib).filter(Meta.yearmo == yearmo)
        return q.count()

    q = dal.session.query(
        Meta.commit_hash, Meta.yearmo, Contrib.name, Contrib.email
    ).join(Contrib)
    return q.count()

def find_file_metrics(dal: DataAccessLayer, yearmo=''):
    if len(yearmo) > 0:
        q = dal.session.query(
            Meta.yearmo,
            func.count(CommitFile.modified_file).label('file_count'),
            func.sum(CommitFile.lines_added).label('lines_added'),
            func.sum(CommitFile.lines_subtracted).label('lines_subtracted')
        ).join(CommitFile).filter(Meta.yearmo == yearmo).\
        group_by(Meta.yearmo)
        return q.first()

    q = dal.session.query(
        Meta.project_name,
        Meta.yearmo,
        func.count(CommitFile.modified_file).label('file_count'),
        func.sum(CommitFile.lines_added).label('lines_added'),
        func.sum(CommitFile.lines_subtracted).label('lines_subtracted')
    ).join(CommitFile).group_by(Meta.project_name)
    return q.first()

def hist_start_yearmo(dal: DataAccessLayer):
    """ Start yearmo """
    q = dal.session.query(
        Meta.yearmo, Author.authored
        ).join(Author)
    return q.first()



""" Handle the data requirements for Assignment 1 

References:
   Assignment 1: "http://janvitek.org/events/NEU/6050/a1.html"
   Git log formatting: "https://git-scm.com/docs/pretty-formats"
"""
import csv
import logging
import os
import subprocess

from okra.proto.assn1_pb2 import Commit, Message, File, Inventory

logger = logging.getLogger(__name__)

def parse_inventory(repopath:str, owner:str, project:str):
    """ Return inventory information for a git repo.

    :param rpath: repository path
    :return: inventory message object
    :rtype: okra.protobuf.assn1_pb2.Inventory
    """
    c = ["git", "log", "-1", "--pretty=%H"]
    res = subprocess.run(c, cwd=repopath, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)

    if res.returncode == 0:
        logger.info("SUCCESS - extracted inventory info")
        hash_val = res.stdout.decode('utf-8', 'ignore').strip()

        inv = Inventory()
        inv.owner = owner
        inv.project = project
        inv.last_hash = hash_val
        return inv
    return None

def parse_commits(rpath: str, chash=""):
    """ Yields a protocol buffer of git commit information.

    commits.csv collects basic information about 
    commits and contains the following columns:

    :param rpath: path to git repository
    :param chash: optional param, retrieve all commits since commit hash
    :return: :class: okra.protobuf.assn1_pb2.Commit
    :rtype: generator, protocol buffer
    """
    if len(chash) == 0:
        c1 = ["git", "log",
              "--pretty=%H^|^%an^|^%ae^|^%aI^|^%cn^|^%ce^|^%cI"]
    else:
        c1 = ["git", "log",
              "--pretty=%H^|^%an^|^%ae^|^%aI^|^%cn^|^%ce^|^%cI",
              "{}..HEAD".format(chash)]

    res = subprocess.run(c1, cwd=rpath, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)

    if res.returncode == 0:
        logger.info("SUCCESS -- extracted git commit info")

        rows = res.stdout.splitlines()

        for row_num, row in enumerate(rows):

            items = row.decode('utf-8', 'ignore').split("^|^")

            if len(items) == 7:
                commit = Commit()

                commit.hash_val = items[0]
                commit.author = items[1]
                commit.author_email = items[2]
                commit.author_timestamp = items[3]
                commit.committer = items[4]
                commit.committer_email = items[5]
                commit.committer_timestamp = items[6]
                
                yield commit

            elif len(items) == 1 and row_num == 0:
                pass # blank line for row 0
            else:
                logger.error("Issue with row {}, repo '{}'".format(row_num, rpath))
                logger.error(row)
    else:
        logger.error("FAIL -- unable to extract git commits info")

def write_line_commits(parsed_commits):
    """ Generate a line for each git commit message. """
    for commit in parsed_commits:

        row = [
            commit.hash_val,
            commit.author,
            commit.author_email,
            commit.author_timestamp,
            commit.committer,
            commit.committer_email,
            commit.committer_timestamp,
        ]
        yield row

def parse_messages(rpath: str, chash=""):
    """ Yields a protocol buffer of a git commit message.
    
    messages.csv collects commit messages and their 
    subject.

    :param rpath: path to git repository
    :param chash: optional param, retrieve all commits since commit hash
    :return: :class: okra.protobuf.assn1_pb2.Message
    :rtype: generator, protocol buffer
    """
    if len(chash) == 0:
        c1 = ["git", "log",
              "--pretty=^^!^^%H^|^%s^|^%b^|^%aI"]
    else:
        c1 = ["git", "log",
              "--pretty=^^!^^%H^|^%s^|^%b^|^%aI",
              "{}..HEAD".format(chash)]
        
    res = subprocess.run(c1, cwd=rpath, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)

    if res.returncode == 0:
        logger.info("SUCCESS -- extracted messages_csv info")

        rows = res.stdout.decode('utf-8', 'ignore').split("^^!^^")

        for row_num, row in enumerate(rows):

            items = row.split("^|^")

            if len(items) == 4:

                message = Message()

                message.hash_val = items[0]
                message.subject = items[1]
                message.message_body = items[2]
                message.timestamp = items[3].strip()

                yield message
             
            elif len(items) == 1 and row_num == 0:
                pass # blank line for row 0   
            else:
                logger.error("Issue with row {}, repo '{}'".format(row_num, rpath))
                logger.error(row)
    else:
        logger.error("FAIL -- unable to extract messages_csv")

def write_line_messages(parsed_messages):
    """ Generate a line for each git commit message. """

    for message in parsed_messages:

        row = [
            message.hash_val,
            message.subject,
            message.message_body,
        ]
        yield row

def parse_committed_files(rpath: str, chash=''):
    """ Parse file format from git log tool. 

    :param rpath: path to git repository
    :param chash: optional param, retrieve all commits since commit hash
    :return: :class: okra.protobuf.assn1_pb2.Message
    :rtype: generator, protocol buffer
    """
    if len(chash) == 0:
        c1 = ["git", "log",
              '--pretty=^|^%n%H',
              '--numstat']
    else:
        c1 = ["git", "log",
              '--pretty=^|^%n%H',
              '--numstat',
              "{}..HEAD".format(chash)]

    res = subprocess.run(c1, cwd=rpath, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)

    if res.returncode != 0:
        logger.error("Unable find file info: {}".format(rpath))        
    
    items = res.stdout.decode('utf-8','ignore').split("^|^")

    for row_num, row in enumerate(items):

        grp = row.strip().splitlines()
        count = 0

        if len(grp) > 0:

            commit_hash = grp[0]

            for file_item in grp[1:]:
                props = file_item.split()

                if len(props) == 0:
                    continue

                finfo = File()
                finfo.hash_val = commit_hash
                finfo.added = props[0]
                finfo.deleted = props[1]
                finfo.file_path = props[2]

                yield finfo




# -*- coding: utf-8 -*-
# pytype: skip-file
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/assn1.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='proto/assn1.proto',
  package='assn1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x11proto/assn1.proto\x12\x05\x61ssn1\"\xa3\x01\n\x06\x43ommit\x12\x10\n\x08hash_val\x18\x01 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x02 \x01(\t\x12\x14\n\x0c\x61uthor_email\x18\x03 \x01(\t\x12\x18\n\x10\x61uthor_timestamp\x18\x04 \x01(\t\x12\x11\n\tcommitter\x18\x05 \x01(\t\x12\x17\n\x0f\x63ommitter_email\x18\x06 \x01(\t\x12\x1b\n\x13\x63ommitter_timestamp\x18\x07 \x01(\t\"U\n\x07Message\x12\x10\n\x08hash_val\x18\x01 \x01(\t\x12\x0f\n\x07subject\x18\x02 \x01(\t\x12\x14\n\x0cmessage_body\x18\x03 \x01(\t\x12\x11\n\ttimestamp\x18\x04 \x01(\t\"K\n\x04\x46ile\x12\x10\n\x08hash_val\x18\x01 \x01(\t\x12\x11\n\tfile_path\x18\x02 \x01(\t\x12\r\n\x05\x61\x64\x64\x65\x64\x18\x03 \x01(\t\x12\x0f\n\x07\x64\x65leted\x18\x04 \x01(\t\">\n\tInventory\x12\r\n\x05owner\x18\x01 \x01(\t\x12\x0f\n\x07project\x18\x02 \x01(\t\x12\x11\n\tlast_hash\x18\x03 \x01(\tb\x06proto3')
)




_COMMIT = _descriptor.Descriptor(
  name='Commit',
  full_name='assn1.Commit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash_val', full_name='assn1.Commit.hash_val', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='author', full_name='assn1.Commit.author', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='author_email', full_name='assn1.Commit.author_email', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='author_timestamp', full_name='assn1.Commit.author_timestamp', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='committer', full_name='assn1.Commit.committer', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='committer_email', full_name='assn1.Commit.committer_email', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='committer_timestamp', full_name='assn1.Commit.committer_timestamp', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29,
  serialized_end=192,
)


_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='assn1.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash_val', full_name='assn1.Message.hash_val', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='subject', full_name='assn1.Message.subject', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message_body', full_name='assn1.Message.message_body', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='assn1.Message.timestamp', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=194,
  serialized_end=279,
)


_FILE = _descriptor.Descriptor(
  name='File',
  full_name='assn1.File',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash_val', full_name='assn1.File.hash_val', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_path', full_name='assn1.File.file_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='added', full_name='assn1.File.added', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deleted', full_name='assn1.File.deleted', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=281,
  serialized_end=356,
)


_INVENTORY = _descriptor.Descriptor(
  name='Inventory',
  full_name='assn1.Inventory',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='owner', full_name='assn1.Inventory.owner', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='project', full_name='assn1.Inventory.project', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='last_hash', full_name='assn1.Inventory.last_hash', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=358,
  serialized_end=420,
)

DESCRIPTOR.message_types_by_name['Commit'] = _COMMIT
DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
DESCRIPTOR.message_types_by_name['File'] = _FILE
DESCRIPTOR.message_types_by_name['Inventory'] = _INVENTORY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Commit = _reflection.GeneratedProtocolMessageType('Commit', (_message.Message,), dict(
  DESCRIPTOR = _COMMIT,
  __module__ = 'proto.assn1_pb2'
  # @@protoc_insertion_point(class_scope:assn1.Commit)
  ))
_sym_db.RegisterMessage(Commit)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), dict(
  DESCRIPTOR = _MESSAGE,
  __module__ = 'proto.assn1_pb2'
  # @@protoc_insertion_point(class_scope:assn1.Message)
  ))
_sym_db.RegisterMessage(Message)

File = _reflection.GeneratedProtocolMessageType('File', (_message.Message,), dict(
  DESCRIPTOR = _FILE,
  __module__ = 'proto.assn1_pb2'
  # @@protoc_insertion_point(class_scope:assn1.File)
  ))
_sym_db.RegisterMessage(File)

Inventory = _reflection.GeneratedProtocolMessageType('Inventory', (_message.Message,), dict(
  DESCRIPTOR = _INVENTORY,
  __module__ = 'proto.assn1_pb2'
  # @@protoc_insertion_point(class_scope:assn1.Inventory)
  ))
_sym_db.RegisterMessage(Inventory)


# @@protoc_insertion_point(module_scope)



""" Command Line Interface utilities for Okra """
import click

from okra.repo_mgmt import clone_or_fetch_repo
from okra.populate_db import persist_repo

@click.group()
def cli():
    pass

@cli.command(name='upsert')
@click.argument('owner')
@click.argument('project')
@click.argument('dburl')
@click.argument('repopath')
@click.argument('repourl')
@click.option('--buffer_size', default=int(1e4), help="number of commits before db write")
def upsert_repo(owner, project, dburl, repopath, repourl, buffer_size):
    clone_or_fetch_repo(repopath, repourl)
    persist_repo(owner, project, dburl, repopath, buffer_size)



""" GitHub Repo Managment

Related to downloading and updating GitHub repos. See
the 'assn1' script in bin/assn1 to handle get and update
features.
"""
import logging
import os
import subprocess
from urllib.parse import urljoin

from okra.error_handling import DirectoryNotCreatedError, NetworkError

logger = logging.getLogger(__name__)

def update_repo(repopath):
    c1 = ["git", "fetch"]
    res = subprocess.run(c1, cwd=repopath, stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE)
    if res.returncode == 0:
        return True
    return False

def clone_repo(repo_url, repo_path):
    try:
        res = subprocess.run(["git", "clone", repo_url, repo_path],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if res.returncode == 0 and os.path.exists(repo_path):
            return True
        raise NetworkError(
                expression=res.stderr,
                message = "Issue with git clone"
        )
    except Exception as exc:
        raise exc

def clone_or_fetch_repo(repo_path, repo_url):
    if os.path.exists(repo_path):
        return update_repo(repo_path)
    return clone_repo(repo_url, repo_path)

def compress_repo(repo_name: str, cache: str, repo_comp: str) -> bool:
    """ Compress repo for upload.

    :param repo_name: git repo name with owner included; 
                      tensorflow/tensorflow
    :param dirpath: directory path to git repo
    :return: creates a compressed file of github repo
    :rtype: True if git repo successfully compressed
    """
    c1 = ["tar", "-zcf", repo_comp, repo_name]
    res = subprocess.run(c1, cwd=cache, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)

    if res.returncode == 0:
        return True
    else:
        logger.error(res.stderr)
        return False

def decompress_repo(repo_comp: str, cache) -> bool:
    """ Decompress repo to a directory.

    :param repo_name: git repo name with owner included; 
                      tensorflow/tensorflow
    :param dirpath: directory path to place uncompressed 
                    file with repo owner
    :param filepath: path to file to be decompressed
    :return: Uncompresses file and writes 
             'git_owner_name/git_repo_name' to the 
             specified directory.
    :rtype: boolean
    :raises: :class:`okra.error_handling.DirectoryNotCreatedError`
    """
    c2 = ["tar", "-zxf", repo_comp, "-C", cache]
    res = subprocess.run(c2, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    if res.returncode == 0:
        return True
    else:
        logger.error(res.stderr)
        return False



"""
Database Schemata

http://janvitek.org/events/NEU/6050/a2.html
"""
import io
from contextlib import redirect_stdout
from sqlalchemy import create_engine
from sqlalchemy import (Table, Column, Integer, String, MetaData,
                        ForeignKey, DateTime)

def config_assn2_schema():
    """ Configure the database schema for assignment 2 """
    metadata = MetaData()

    commit_meta = Table('commit_meta', metadata,
                        Column('id', Integer, primary_key=True, nullable=False),
                       Column('owner_name', String, nullable=False),
                       Column('project_name', String, nullable=False),
                       Column('commit_hash', String,
                              ForeignKey("commit_info.commit_hash"),
                              ForeignKey("commit_author.commit_hash"),
                              ForeignKey("commit_contribs.commit_hash"),
                              ForeignKey("commit_file.commit_hash"),
                              nullable=False, index=True, unique=True),
    )

    commit_author = Table('commit_author', metadata,
                          Column('id', Integer, primary_key=True),
                          Column('commit_hash', String, nullable=False,
                                 index=True, unique=True),
                          Column('author_name', String, nullable=False,
                                 index=True),
                          Column('author_email', String, nullable=True),
                          Column('author_datetime', DateTime, nullable=False),
    )

    commit_contribs = Table('commit_contribs', metadata,
                            Column('id', Integer, primary_key=True),
                            Column('commit_hash', String, nullable=False,
                                   index=True, unique=True),
                            Column('contrib_name', String, nullable=False,
                                   index=True),
                            Column('contrib_email', String, nullable=True),
                            Column('contrib_datetime', DateTime,
                                   nullable=False),
    )

    commit_file = Table('commit_file', metadata,
                        Column('id', Integer, primary_key=True),
                        Column('commit_hash', String, nullable=False,
                               index=True, unique=True),
                        Column('modified_file', String, nullable=False),
                        Column('modified_add_lines', Integer, nullable=False),
                        Column('modified_subtract_lines', Integer,
                               nullable=False),
    )

    commit_info = Table('commit_info', metadata,
                        Column('id', Integer, primary_key=True),
                        Column('commit_hash', String, nullable=False,
                               index=True, unique=True),
                        Column('subject', String, nullable=False),
                        Column('message', String, nullable=False),
    )

    return metadata

def metadata_tosql(metadata: MetaData, db_url: str):
    """ Convert Database metadata into SQL statements. 
    
    Generate SQL for any database supported by SQLAlchemy.
    No commands are executed on a database by this function
    so the db_url can be a mock.

    :param metadata: MetaData object from sqlalchemy containing db schema
    :param db_url: database url, can be a mock
    :return: SQL database schema for db_url database type
    :rtype: string
    """

    def metadata_dump(sql, *multiparams, **params):
        """ print or write to log or file etc """
        print(sql.compile(dialect=engine.dialect))

    engine = create_engine(db_url, strategy='mock', executor=metadata_dump)

    f = io.StringIO()
    with redirect_stdout(f):
        metadata.create_all(engine)
    out = f.getvalue()
    return out
    



""" Turns out that joins are less than ideal in Spark. 

Instead of using a central database, we'll solve the problem in a
distributed way. Each function returns an immutable type. We'll just
use one server with like 32 cores or whatever. Need to throw out the
parquet files and use the SQLite databases.

Reference:
    tbonza/EDS19/project/PredictRepoHealth.ipynb
"""
from collections import defaultdict
from datetime import datetime
import logging
from multiprocessing import Pool
import os
import re

import numpy as np
import pandas as pd

from okra.models import Meta, Author, Contrib, CommitFile, Info
from okra.models import DataAccessLayer


logger = logging.getLogger(__name__)


def run_distributed_pool(n_cores:int, func, work: list):
    """ Run distributed pool across n_cores

    :param n_cores: number of cores in available in server
    :param func: immutable function to execute work
    :param work: input for immutable function
    :return: func output from work
    :rtype: meta
    """
    with Pool(n_cores) as p:
        result = p.map(func, work)
    return result

def create_working_table(dal: DataAccessLayer):
    """ Create a base table to derive all other features and aggregations

    :param dal: okra.models.DataAccessLayer, session must be instantiated
    :return: base table for analysis
    :rtype: pd.DataFrame
    """
    if dal.session is None:
        raise TypeError("Session must be instantiated")

    q = dal.session.query(
        Meta.commit_hash, Meta.owner_name, Meta.project_name,
        CommitFile.file_id, CommitFile.modified_file,
        CommitFile.lines_added, CommitFile.lines_deleted,
        Author.name, Author.email, Author.authored)\
        .join(Author).join(CommitFile)

    items = []
    for item in q.all():
        r = {
            'commit_hash': item.commit_hash,
            'owner_name': item.owner_name,
            'project_name': item.project_name,
            'file_id': item.file_id,
            'modified_file': item.modified_file,
            'lines_added': item.lines_added,
            'lines_deleted': item.lines_deleted,
            'name': item.name,
            'email': item.name,
            'authored': item.authored,
        }
        items.append(r)

    df = pd.DataFrame(items)
    df['ts'] = pd.to_datetime(df.authored)

    return df

def num_authors(df, period: str):
    """ Number of authors in a given time period. 

    :param df: pd.DataFrame, base table
    :param period: str, 'Y', year
    :return: df aggregated by period
    :rtype: pd.DataFrame
    """
    per = df.ts.dt.to_period(period)
    
    cols = ['name','owner_name', 'project_name']
    grp = [per, 'owner_name', 'project_name']
    result = df[cols].groupby(grp).count()
    result.columns = ['num_authors']
    result = result.reset_index()
    return result

def num_mentors(df, period:str, subperiod:str, k:int):
    """ 
    Number of authors in a larger time period who have also
    made commits in k number of smaller time periods. 
    
    For example, number of authors in a year who have also 
    committed changes in each month of that year.

    :param df: pd.DataFrame, base table
    :param period: 'Y', year
    :param subperiod: 'M', month
    :param k: int, authors exist at least k subperiods
    :return: df aggregated by period
    :rtype: pd.DataFrame
    """
    subper = df.ts.dt.to_period(subperiod)
    subcols = ['owner_name', 'project_name',
               'name', 'email']
    subgrp = [subper, 'owner_name', 'project_name', 'name', 'email']
    ok = df[subcols].groupby(subgrp).count()
    ok = ok.reset_index()

    if period == 'Y':

        ok['year'] = ok.ts.apply(lambda x: x.year)
        cols = ['year', 'owner_name', 'project_name', 'name', 'email']
        grp = ['year','owner_name','project_name', 'name']
        ok = ok[cols].groupby(grp).count()

        ok = ok.reset_index()
        ok = ok[['year', 'owner_name', 'project_name', 'email']]
        ok.columns = ['ts', 'owner_name', 'project_name', 'mentor_count']
        result = ok[ok.mentor_count >= k].groupby(['ts', 'owner_name',
                                                   'project_name']).count()
        
        result = result.reset_index()
        return result

    else:
        raise Exception("Period '{}' not found".format(period))

    return df


def num_orgs(df, period):
    """ Number of organizations commiting to a repo in a time period

    Organizations are found by extracting the domain of email 
    addresses used by authors.

    :param df: pd.DataFrame, base table
    :param period: str, 'Y' year time period
    :return: df aggregated by period
    :rtype: pd.DataFrame
    """
    email_suffix = ['com','org','ru','edu','dk','id.au','ac',
                    'de','uk','cz','fr','jp','il','me','net',
                    'eu', 'pw', 'ch','cn','io','nu','it','ai',
                    'fi', 'info', 'sk', 'ie', 'ca', 'at', 'pl',
                    'hu', 'nl', 'works', 'hr', 'se','no', 'blue',
                    '(none)', 'lt', 'cc', 'si', 'la', 'us', 'gr',
                    'De','br', 'ro', 'li', 'gov', 'pt', 'is', 'sg',
                    'vin', 'in', 'be', 'bzh', 'pm', 'xyz', 'fun',
                    'es', 'systems', 'cx', 'tw','cl', 'lv', 'ne',
                    'co', 'st', 'ma', 'ws', 'su', 'ORG', 'tk', 'cu',
                    'COM', 'kr', 'vpc', 'ip', 'danix', 'Com', 're',
                    'EDU']

    pattern = '[A-Za-z\-\._0-9\+]+@(.*)?.({})'.format('|'.join(email_suffix))
    pat = re.compile(pattern)

    df['org_name'] = df.email.apply(lambda x: pat.match(x).group(1)
                                    if pat.match(x) else np.NaN)
    
    per = df.ts.dt.to_period(period)
    cols = ['owner_name', 'project_name', 'org_name']

    result = df[cols].groupby([per, 'owner_name', 'project_name']).count()
    result = result.reset_index()

    return result

def create_features_target(df, k=6):
    """ Create dataset from train/test/val and Y """

    ## Features
    
    X_1 = num_authors(df, 'Y')
    X_2 = num_mentors(df, 'Y', 'M', k)
    X_3 = num_orgs(df, 'Y')

    X_13 = pd.merge(X_1, X_3)
    X_13.ts  = X_13.ts.apply(lambda x: x.year)
    assert X_13.shape[0] == X_1.shape[0]

    cols = ['ts','owner_name', 'project_name', 'mentor_count']
    join_cols = ['ts','owner_name', 'project_name']
    df_features = X_13.merge(X_2[cols], how='left', on=join_cols)

    logger.debug('X_2 head: {}'.format(X_2.head()))
    logger.debug('df_features head: {}'.format(df_features.head()))

    ## Target

    per = df.ts.dt.to_period('Y')
    cols = ['owner_name', 'project_name', 'lines_added', 'lines_deleted']
    grp = [per, 'owner_name', 'project_name']
    df_target = df[cols].groupby(grp).sum()

    df_target['velocity'] = df_target.lines_added - \
        df_target.lines_deleted

    ## Reset indices
    
    df_features = df_features.reset_index()
    df_target = df_target.reset_index()

    ## Recast pandas specific types

    df_target.ts = df_target.ts.apply(lambda x: x.year)

    return df_features, df_target

def write_features_target(dbinfo: tuple, k=6) -> bool:
    """ Write out features and target dataframes to parquet. 
    
    :param dburl: SQLite database url
    :return: write out feature, target dataframes as parquet
    :rtype: bool
    """
    try:
        dburl, repo_name, cache = dbinfo
        
        dal = DataAccessLayer(dburl)
        dal.connect()
        dal.session = dal.Session()

        df = create_working_table(dal)
        df_features, df_target = create_features_target(df, k)
        
        logger.debug("target dtypes: {}".format(df_target.dtypes))
        logger.debug("features dtypes: {}".format(df_features.dtypes))

        logger.debug("target shape: {}".format(df_target.shape))
        logger.debug("features shape: {}".format(df_features.shape))

        feature_path = "{}features_{}.parquet".format(cache, repo_name)
        target_path = "{}target_{}.parquet".format(cache, repo_name)

        df_features.to_parquet(feature_path, 'pyarrow')
        df_target.to_parquet(target_path, 'pyarrow')
        logger.info("Wrote features: {}".format(feature_path))
        logger.info("Wrote target: {}".format(target_path))
        return True

    except Exception as exc:
        logger.exception(exc)
        return False

def getwork_dbinfo(cache: str):
    """ Defines work for immutable function, write_features_target 

    :param cache: str, full path to cache
    :return: dbinfo, (dburl, repo_name, cache)
    :rtype: list of tuples
    """
    work = []
    try:
        dbs = [i for i in os.listdir(cache) if i[-3:] == ".db"]

        for db in dbs:

            dburl = "sqlite:///" + cache + db
            repo_name = db[:-3].replace("__REPODB__", "__LOGDF__")

            dbinfo = (dburl, repo_name, cache)
            work.append(dbinfo)

        return work

    except Exception as exc:
        logger.exception(exc)
        return []

def consolidate_features_target(cache: str, repo_id: str, report=100):
    """ Consolidates distributed computations in cache

    We're going to have n feature tables, n target tables. We
    need to consolidate them into one feature and one target
    table. This is the 'reduce' part of a 'map/reduce' pattern. 
    To avoid confusion with Hadoop, let's call this a 
    'hack/reduce' pattern.
    
    :param cache: str, full path to cache
    :return: X_features, X_target written to disk
    :rtype: None
    """
    date_today = datetime.now().strftime("%Y%m%d")        
    df_names = [i for i in os.listdir(cache) if repo_id in i]

    feature_names = [i for i in df_names if 'features_' in i]
    target_names = [i for i in df_names if 'target_' in i]

    first = True
    fcount = 0
    for feature_name in feature_names:

        fpath = cache + feature_name
        if first:
            featdf = pd.read_parquet(fpath, 'pyarrow')
            first = False

        else:
            fdf = pd.read_parquet(fpath, 'pyarrow')
            featdf = featdf.append(fdf)

        if fcount % report == 0:
            logger.info("Processed {} feature dataframes".format(fcount))
        fcount += 1

    feat_path = '{}df_features_{}.parquet'.format(cache, date_today)
    featdf.to_parquet(feat_path, 'pyarrow')
    logger.info('Wrote df_features, {}, to {}'.\
                format(featdf.shape,feat_path))

    first = True
    tcount = 0
    for target_name in target_names:

        fpath = cache + target_name
        if first:
            targdf = pd.read_parquet(fpath, 'pyarrow')
            first = False

        else:
            tdf = pd.read_parquet(fpath, 'pyarrow')
            targdf = targdf.append(tdf)

        if tcount % report == 0:
            logger.info("Processed {} target dataframes".format(tcount))
        tcount += 1

    targ_path = '{}df_target_{}.parquet'.format(cache, date_today)
    targdf.to_parquet(targ_path, 'pyarrow')
    logger.info('Wrote df_target, {}, to {}'.\
                format(targdf.shape, targ_path))

    
    



""" Custom errors we can expect. 

References:
  https://docs.python.org/3/tutorial/errors.html
"""

class Error(Exception):
    """ Base class for exceptions in okra """
    pass

class NetworkError(Error):
    """ Exception raised for errors related to network requests

    :param expression: input expression in which the error occurred
    :param message: explanation of error
    """

    def __init__(self, expression, message):
        self.expression = expression
        self.message = message

class MissingEnvironmentVariableError(Error):
    """ Exception raised when mandatory enviroment variable is missing.

    :param expression: input expression in which error occurred
    :param message: explanation of error
    """

    def __init__(self, expression, message):
        self.expression = expression
        self.message = message

class DirectoryNotCreatedError(Error):
    """ Exception raised when directory unable to be created.

    :param expression: input expression in which error occurred
    :param message: explanation of error
    """

    def __init__(self, expression, message):
        self.expression = expression
        self.message = message


    



import os
import logging

def enable_log(log_name):
    """ Enable logs written to file """
    logging.basicConfig(filename= log_name,
                        level=logging.INFO,
                        format='%(asctime)s %(levelname)s %(message)s')

def enable_cloud_log(level='INFO'):
    """ Enable logs using default StreamHandler """
    levels = {'INFO': logging.INFO, 'DEBUG': logging.DEBUG}
    logging.basicConfig(level=levels[level],
                        format='%(asctime)s %(levelname)s %(message)s')




""" Generate tables for the Okra-API management command


Metrics are targeting the 'proto/okra_api.proto' interface.
"""
from datetime import datetime
import os
import logging
from urllib.parse import urljoin

from sqlalchemy import func

from okra.assn4 import (get_truck_factor_by_project,
                        total_number_of_files_by_project,
                        total_number_of_contributors_by_project)
from okra.bus_factor import get_bus_factor
from okra.models import DataAccessLayer, Author, Meta, Contrib, CommitFile
from okra.proto import okra_api_pb2
from okra.repository_metrics import (repo_info, iso_date_aggregation,
                                     find_author_count,
                                     find_contributor_count,
                                     find_file_metrics,
                                     hist_start_yearmo)

logger = logging.getLogger(__name__)

def date_toiso(datetime, timespec='minutes'):
    return datetime.isoformat(timespec=timespec)

def msg_iso_date_aggregation(msg, item, status: str, yearmo: str):
    """ Compute IsoDateAggregation message """

    isodt = msg.isodates.add()
    isodt.yearmo = yearmo
    isodt.commit_hash = item.commit_hash

    isoyr, isowk, isody = item.contributed.isocalendar()

    isodt.iso_week = isowk
    isodt.iso_year = isoyr
    isodt.status = status

    return msg  

def msg_repository_info(dal: DataAccessLayer, repo_id: str, yearmo: str):
    """ Compute RepositoryInfo message 

    Note that default behavior for first/last msg_iso_date_aggregation()
    is to use the same msg item for IsoDateAggregation if only one msg
    item exists for a given yearmo. An empty message except for repo_id
    and yearmo will be returned if no commits exist for a given yearmo.
    """

    msg = okra_api_pb2.RepositoryInfo()

    qres = repo_info(dal, yearmo)

    if len(qres) == 0:
        msg.repo_id = repo_id
        msg.yearmo = yearmo
        return msg

    msg.repo_id = repo_id
    msg.yearmo = yearmo

    # Set up iso dates (first, last)

    first = qres[0]
    last = qres[-1]

    msg = msg_iso_date_aggregation(msg, item=first, status='first',
                                   yearmo=yearmo)
    msg = msg_iso_date_aggregation(msg, item=last, status='last',
                                   yearmo=yearmo)
    return msg

def msg_repository_metric(dal: DataAccessLayer, repo_id: str, yearmo: str):
    
    msg = okra_api_pb2.RepositoryMetric()

    msg.repo_id = repo_id
    msg.yearmo = yearmo

    # IsoDataAggregation

    qres1 = iso_date_aggregation(dal, yearmo)

    if len(qres1) == 0:
        return msg

    first = qres1[0]
    last = qres1[-1]

    msg = msg_iso_date_aggregation(msg, first, status='first', yearmo=yearmo)
    msg = msg_iso_date_aggregation(msg, last, status='last', yearmo=yearmo)

    # Author count

    msg.author_count = find_author_count(dal, yearmo)
    
    # Contributor count

    msg.contrib_count = find_contributor_count(dal, yearmo)

    # File level metrics

    qres4 = find_file_metrics(dal, yearmo)
    msg.file_count = qres4.file_count
    msg.lines_added = qres4.lines_added
    msg.lines_subtracted = qres4.lines_subtracted

    # Truck factor

    member_count, members = get_bus_factor(dal, yearmo)
    msg.truck_factor = member_count

    return msg

def msg_repo_history_metric(dal: DataAccessLayer, repo_id: str, yearmo: str):

    msg = okra_api_pb2.RepositoryHistoryMetric()

    msg.repo_id = repo_id
    msg.current_yearmo = yearmo

    # IsoDataAggregation
        
    qres1 = iso_date_aggregation(dal)

    if len(qres1) == 0:
        return msg

    first = qres1[0]
    last = qres1[-1]

    msg = msg_iso_date_aggregation(msg, first, status='first',
                                   yearmo=first.yearmo)
    msg = msg_iso_date_aggregation(msg, last, status='last',
                                   yearmo=last.yearmo)

    # Find start yearmo

    msg.start_yearmo = first.yearmo

    # Find last commit info

    msg.last_commit_yearmo = last.yearmo
    msg.last_commit_hash = last.commit_hash

    # Total lines added/subtracted

    fimet = find_file_metrics(dal)
    msg.total_lines_added = fimet.lines_added
    msg.total_lines_subtracted = fimet.lines_subtracted

    # Total truck factor

    member_count, members = get_bus_factor(dal)
    msg.total_truck_factor = member_count

    # Total days in project

    time_elapsed = last.contributed - first.contributed
    msg.total_days = time_elapsed.days

    return msg



# -*- coding: utf-8 -*-
# pytype: skip-file
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/okra_api.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='proto/okra_api.proto',
  package='okra.v1',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x14proto/okra_api.proto\x12\x07okra.v1\"m\n\x12IsoDateAggregation\x12\x0e\n\x06yearmo\x18\x01 \x01(\t\x12\x13\n\x0b\x63ommit_hash\x18\x02 \x01(\t\x12\x10\n\x08iso_week\x18\x03 \x01(\x05\x12\x10\n\x08iso_year\x18\x04 \x01(\x05\x12\x0e\n\x06status\x18\x05 \x01(\t\"\xa8\x02\n\x17RepositoryHistoryMetric\x12\x0f\n\x07repo_id\x18\x01 \x01(\t\x12\x14\n\x0cstart_yearmo\x18\x02 \x01(\t\x12\x16\n\x0e\x63urrent_yearmo\x18\x03 \x01(\t\x12-\n\x08isodates\x18\x04 \x03(\x0b\x32\x1b.okra.v1.IsoDateAggregation\x12\x1a\n\x12last_commit_yearmo\x18\x05 \x01(\t\x12\x18\n\x10last_commit_hash\x18\x06 \x01(\t\x12\x19\n\x11total_lines_added\x18\x07 \x01(\x05\x12\x1e\n\x16total_lines_subtracted\x18\x08 \x01(\x05\x12\x1a\n\x12total_truck_factor\x18\t \x01(\x05\x12\x12\n\ntotal_days\x18\n \x01(\x02\"\xe8\x01\n\x10RepositoryMetric\x12\x0f\n\x07repo_id\x18\x01 \x01(\t\x12\x0e\n\x06yearmo\x18\x02 \x01(\t\x12-\n\x08isodates\x18\x03 \x03(\x0b\x32\x1b.okra.v1.IsoDateAggregation\x12\x14\n\x0c\x61uthor_count\x18\x04 \x01(\x05\x12\x15\n\rcontrib_count\x18\x05 \x01(\x05\x12\x12\n\nfile_count\x18\x06 \x01(\x05\x12\x13\n\x0blines_added\x18\x07 \x01(\x05\x12\x18\n\x10lines_subtracted\x18\x08 \x01(\x05\x12\x14\n\x0ctruck_factor\x18\t \x01(\x05\"`\n\x0eRepositoryInfo\x12\x0f\n\x07repo_id\x18\x01 \x01(\t\x12\x0e\n\x06yearmo\x18\x02 \x01(\t\x12-\n\x08isodates\x18\x03 \x03(\x0b\x32\x1b.okra.v1.IsoDateAggregationb\x06proto3')
)




_ISODATEAGGREGATION = _descriptor.Descriptor(
  name='IsoDateAggregation',
  full_name='okra.v1.IsoDateAggregation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='yearmo', full_name='okra.v1.IsoDateAggregation.yearmo', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='commit_hash', full_name='okra.v1.IsoDateAggregation.commit_hash', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='iso_week', full_name='okra.v1.IsoDateAggregation.iso_week', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='iso_year', full_name='okra.v1.IsoDateAggregation.iso_year', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='status', full_name='okra.v1.IsoDateAggregation.status', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33,
  serialized_end=142,
)


_REPOSITORYHISTORYMETRIC = _descriptor.Descriptor(
  name='RepositoryHistoryMetric',
  full_name='okra.v1.RepositoryHistoryMetric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='repo_id', full_name='okra.v1.RepositoryHistoryMetric.repo_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_yearmo', full_name='okra.v1.RepositoryHistoryMetric.start_yearmo', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='current_yearmo', full_name='okra.v1.RepositoryHistoryMetric.current_yearmo', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='isodates', full_name='okra.v1.RepositoryHistoryMetric.isodates', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='last_commit_yearmo', full_name='okra.v1.RepositoryHistoryMetric.last_commit_yearmo', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='last_commit_hash', full_name='okra.v1.RepositoryHistoryMetric.last_commit_hash', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_lines_added', full_name='okra.v1.RepositoryHistoryMetric.total_lines_added', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_lines_subtracted', full_name='okra.v1.RepositoryHistoryMetric.total_lines_subtracted', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_truck_factor', full_name='okra.v1.RepositoryHistoryMetric.total_truck_factor', index=8,
      number=9, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_days', full_name='okra.v1.RepositoryHistoryMetric.total_days', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=145,
  serialized_end=441,
)


_REPOSITORYMETRIC = _descriptor.Descriptor(
  name='RepositoryMetric',
  full_name='okra.v1.RepositoryMetric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='repo_id', full_name='okra.v1.RepositoryMetric.repo_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='yearmo', full_name='okra.v1.RepositoryMetric.yearmo', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='isodates', full_name='okra.v1.RepositoryMetric.isodates', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='author_count', full_name='okra.v1.RepositoryMetric.author_count', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contrib_count', full_name='okra.v1.RepositoryMetric.contrib_count', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_count', full_name='okra.v1.RepositoryMetric.file_count', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lines_added', full_name='okra.v1.RepositoryMetric.lines_added', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lines_subtracted', full_name='okra.v1.RepositoryMetric.lines_subtracted', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='truck_factor', full_name='okra.v1.RepositoryMetric.truck_factor', index=8,
      number=9, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=444,
  serialized_end=676,
)


_REPOSITORYINFO = _descriptor.Descriptor(
  name='RepositoryInfo',
  full_name='okra.v1.RepositoryInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='repo_id', full_name='okra.v1.RepositoryInfo.repo_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='yearmo', full_name='okra.v1.RepositoryInfo.yearmo', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='isodates', full_name='okra.v1.RepositoryInfo.isodates', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=678,
  serialized_end=774,
)

_REPOSITORYHISTORYMETRIC.fields_by_name['isodates'].message_type = _ISODATEAGGREGATION
_REPOSITORYMETRIC.fields_by_name['isodates'].message_type = _ISODATEAGGREGATION
_REPOSITORYINFO.fields_by_name['isodates'].message_type = _ISODATEAGGREGATION
DESCRIPTOR.message_types_by_name['IsoDateAggregation'] = _ISODATEAGGREGATION
DESCRIPTOR.message_types_by_name['RepositoryHistoryMetric'] = _REPOSITORYHISTORYMETRIC
DESCRIPTOR.message_types_by_name['RepositoryMetric'] = _REPOSITORYMETRIC
DESCRIPTOR.message_types_by_name['RepositoryInfo'] = _REPOSITORYINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

IsoDateAggregation = _reflection.GeneratedProtocolMessageType('IsoDateAggregation', (_message.Message,), dict(
  DESCRIPTOR = _ISODATEAGGREGATION,
  __module__ = 'proto.okra_api_pb2'
  # @@protoc_insertion_point(class_scope:okra.v1.IsoDateAggregation)
  ))
_sym_db.RegisterMessage(IsoDateAggregation)

RepositoryHistoryMetric = _reflection.GeneratedProtocolMessageType('RepositoryHistoryMetric', (_message.Message,), dict(
  DESCRIPTOR = _REPOSITORYHISTORYMETRIC,
  __module__ = 'proto.okra_api_pb2'
  # @@protoc_insertion_point(class_scope:okra.v1.RepositoryHistoryMetric)
  ))
_sym_db.RegisterMessage(RepositoryHistoryMetric)

RepositoryMetric = _reflection.GeneratedProtocolMessageType('RepositoryMetric', (_message.Message,), dict(
  DESCRIPTOR = _REPOSITORYMETRIC,
  __module__ = 'proto.okra_api_pb2'
  # @@protoc_insertion_point(class_scope:okra.v1.RepositoryMetric)
  ))
_sym_db.RegisterMessage(RepositoryMetric)

RepositoryInfo = _reflection.GeneratedProtocolMessageType('RepositoryInfo', (_message.Message,), dict(
  DESCRIPTOR = _REPOSITORYINFO,
  __module__ = 'proto.okra_api_pb2'
  # @@protoc_insertion_point(class_scope:okra.v1.RepositoryInfo)
  ))
_sym_db.RegisterMessage(RepositoryInfo)


# @@protoc_insertion_point(module_scope)



""" Custom exceptions for Okra. """

class OkraError(Exception):
    """ Base class for exceptions in Okra """
    pass

class UrlJoinError(OkraError):
    """ Exception raised for error in 'urllib.parse.urljoin'

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """
    def __init__(self, expression, message):
        self.expression = expression
        self.message = message




""" A nice approach to to downloading GitHub repos.

Previous approaches to downloading GitHub repositories
were set up in parallel on Kubernetes using Redis. This
approach was considered too aggressive by both GitHub and
Google. This approach is meant to be nice.
"""
import csv
import logging
import os
import time
from urllib.parse import urljoin

from okra.populate_db import persist_repo


# TODO be nice



""" Batch job utilities. """

from okra.exceptions import UrlJoinError



def check_urljoin(cache: str, fname: str, fpath: str) -> bool:
    if fpath == fname:

        exp = 'cache: {}, fname: {}, fpath: {}'.\
            format(cache, fname, fpath)
        raise UrlJoinError(expression=exp, message='incorrect urljoin')

    return True



from setuptools import setup, find_packages

setup(
    name='okra',
    version='2.0.0',
    packages=["okra", "okra/proto"],
    license='MIT',
    author="Tyler Brown",
    author_email="brown.tyler@husky.neu.edu",
    long_description_content_type="text/markdown",
    long_description=open('README.md').read(),
    python_requires=">=3.7",
    install_requires=[
        'click == 7.0',
        'pandas == 1.0.1',
        'protobuf == 3.11.3',
        'sqlalchemy == 1.3.13',
    ],
    zip_safe=False,
    url='https://okrahealth.github.io/',
    project_urls={
        "Bug Tracker": "https://github.com/okrahealth/okra/issues",
        "Documentation": "https://okrahealth.github.io/okra/",
        "Source Code": "https://github.com/okrahealth/okra",
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points={
        'console_scripts': ['okra = okra.cli:cli'],
    }
)



""" Place log information into objects 

This is going to generate items for each specified model object
based on git log commands. 
"""
from datetime import datetime
import logging
import os
import re
from urllib.parse import urljoin

from okra.error_handling import DirectoryNotCreatedError
from okra.models import Meta, Author, Contrib, CommitFile, Info, Inventory
from okra.gitlogs import (parse_commits, parse_messages,
                          parse_committed_files)

logger = logging.getLogger(__name__)

def make_digit(numstr, desc):
    try:
        d = int(numstr)
        return d
    
    except ValueError:
        d = 0
        if numstr.strip() != '-':
            logger.error("ValueError for {}: {}".format(desc, numstr))
        return d

def repo_to_objects(owner:str, project:str, repopath:str, last_commit=""):
    """ Retrieve objects from last commit if exists

    This function is a generator so we can specify a buffer size
    when making commits to the database. Otherwise, the I/O would
    slow things way down.

    :param repo_name: git user/git repo name, 'tbonza/EDS19'
    :param dirpath: path to directory storing repo information
    :param last_commit: string of git commit hash last stored in database
    :return: generator of populated model database objects
    :rtype: sqlalchemy database objects
    """
    if len(last_commit) == 0:
        
        cmts = parse_commits(repopath)
        msgs = parse_messages(repopath)
        fobjs = parse_committed_files(repopath)

    else:
        # retrieve from last commit HEAD
        # need to set 'c1' lists

        cmts = parse_commits(repopath, chash=last_commit)
        msgs = parse_messages(repopath, chash=last_commit)
        fobjs = parse_committed_files(repopath, chash=last_commit)

    # map objects to database objects

    m = re.compile(r"Co-authored-by\:(.*?)<(.*?)>")
    for msg, cmt in zip(msgs, cmts):
        contrib_id = 0

        commit_authored_datetime = datetime.fromisoformat(cmt.committer_timestamp)
        yrmo_cad = str(commit_authored_datetime.year) + "-" + str(commit_authored_datetime.month)

        msg_item = Info(
            commit_hash=msg.hash_val,
            subject=msg.subject,
            message=msg.message_body,
            created=datetime.fromisoformat(msg.timestamp)
        )

        meta_item = Meta(
            commit_hash=msg.hash_val,
            owner_name=owner,
            project_name=project,
            yearmo=yrmo_cad
        )
        
        author_item = Author(
            commit_hash=cmt.hash_val,
            name=cmt.author,
            email=cmt.author_email,
            authored=commit_authored_datetime
        )

        contrib_item = Contrib(
            contrib_id=contrib_id,
            commit_hash=cmt.hash_val,
            name=cmt.committer,
            email=cmt.committer_email,
            contributed=datetime.fromisoformat(cmt.committer_timestamp)
        )        
        yield msg_item
        yield meta_item
        yield author_item
        yield contrib_item

        # some commits will have multiple contributors
        
        if m.match(msg.message_body):

            contrib_id += 1
            for item in m.findall(msg.message_body):

                contrib_item = Contrib(
                    contrib_id=contrib_id,
                    commit_hash=msg.hash_val,
                    name=item[0].strip(),
                    email=item[1].strip(),
                    contributed=datetime.fromisoformat(msg.timestamp)
                )
                yield contrib_item
                contrib_id += 1

    file_id = 0
    for fobj in fobjs:

        cf_item = CommitFile(
            file_id=file_id,
            commit_hash=fobj.hash_val,
            modified_file=fobj.file_path,
            lines_added=make_digit(fobj.added, "fobj.added"),
            lines_subtracted=make_digit(fobj.deleted, "fobj.deleted")
        )
        yield cf_item

        file_id += 1



""" Compute Truck Factor

The truck factor assignment includes several factors. We're just
going to focus on the actual truck factor computation from a database
in this file.

References:
  Assignment: http://janvitek.org/events/NEU/6050/a4.html
  Paper: http://janvitek.org/events/NEU/6050/Ps/truck.pdf

Modified version of okra/assn4.py for yearmo 
"""
from collections import defaultdict

from sqlalchemy import func
from okra.assn4 import author_number_of_files_owned, smallest_owner_set
from okra.models import (Meta, Author, Contrib, CommitFile, Info)

def number_of_files(dal, yearmo=''):
    """ Compute the number of files by project

    If yearmo, computation only applies to the last
    yearmo.

    :param dal: okra.models.DataAccessLayer
    :param yearmo: str, year month
    :return: the number of files in a project
    :rtype: int
    """
    if len(yearmo) > 0:
        q = dal.session.query(
            Meta.yearmo, CommitFile.modified_file
        ).join(CommitFile).filter(Meta.yearmo == yearmo).\
        group_by(CommitFile.modified_file)
        return q.count()

    q = dal.session.query(CommitFile.modified_file).\
        group_by(CommitFile.modified_file)
    return q.count()

def number_of_contributors(dal, yearmo=''):
    """
    :param dal: okra.models.DataAccessLayer
    :param yearmo: str, year month
    :return: the number of contributors in a project
    :rtype: int
    """
    if len(yearmo) > 0:
        q = dal.session.query(
            Meta.yearmo, Contrib.name, Contrib.email
        ).filter(Meta.yearmo == yearmo).join(Contrib).\
        group_by(Contrib.name, Contrib.email)
        return q.count()

    q = dal.session.query(Contrib.name, Contrib.email)
    return q.count()

def max_lines_perfile(res):
    """ Author with max number of lines added per file (owner) """
    
    owner = {}
    for item in res:

        fname = item.modified_file
        if fname in owner:

            if owner[fname].total_lines_added > item.total_lines_added:
                owner[fname] = item

        else:
            owner[fname] = item

    results = list(owner.values())
    return results

def file_author_owner(dal, yearmo=''):
    """ Compute file ownership by each author. 

    :param dal: okra.models.DataAccessLayer
    :param yearmo: str, year month
    :return: list of file owners based on max number of lines written
    :rtype: list of sqlalchemy objects
    """
    if len(yearmo) > 0:
        
        # Number of lines added by each author per file, yearmo filter

        q = dal.session.query(
            Meta.commit_hash, Meta.yearmo, Author.name, Author.email,
            CommitFile.modified_file,
            func.sum(CommitFile.lines_added).label("total_lines_added")
        ).filter(Meta.yearmo == yearmo).\
        join(Author).join(CommitFile).\
        group_by(Author.name, Author.email, CommitFile.modified_file)

        res = q.all()
        return max_lines_perfile(res)

    # Number of lines added by each author per file

    q = dal.session.query(
        Meta.commit_hash, Meta.yearmo, Author.name, Author.email,
        CommitFile.modified_file,
        func.sum(CommitFile.lines_added).label("total_lines_added")
    ).join(Author).join(CommitFile).\
    group_by(Author.name, Author.email, CommitFile.modified_file)
    
    res = q.all()
    return max_lines_perfile(res)

def get_bus_factor(dal, yearmo=''):
    """ Get the 'bus factor'. 

    1. For each project, and each file, compute how many lines were 
       added by each unique user.
    2. For each project, and each file, find which user created the file.
    3. Given the above two results compute the ownership of each file.
    4. For each project, and each file pick an owner.
    5. For each project, rank the users by the number of files they own.
    6. Given all of the above compute the Truck Factor as the smallest set 
       of users such that they own more than half of the files in the project

    :param dal: okra.models.DataAccessLayer
    :param yearmo: str, year month
    :return: Truck factor score for a GitHub project, Truck set members
    :rtype: tuple (int, list)
    """
    total_files = number_of_files(dal, yearmo)
    owners = file_author_owner(dal, yearmo)
    authors = author_number_of_files_owned(owners)
    member_count, members = smallest_owner_set(authors, total_files)

    return (member_count, members)
